@@ -59,7 +59,7 @@
 static void usage (void)
 {
 	fprintf (stderr, "\
-paexec - parallel executor\n\
+paexec -- parallel executor\n\
          that distributes tasks over CPUs or machines in a network.\n\
 usage: paexec    [OPTIONS]\n\
        paexec -C [OPTIONS] cmd [args...]\n\
@@ -70,10 +70,15 @@ OPTIONS:\n\
   -n <+num>        number of subprocesses to run\n\
   -n <nodes>       list of nodes separated by space character\n\
   -n <:filename>   filename containing a list of nodes, one node per line\n\
-  -c <command>     path to a command\n\
-  -t <trans>       path to a transport program\n\
+  -c <command>     set a command\n\
+  -C               use free arguments as a command and its arguments\n\
+  -t <transport>   set a transport program\n\
+\n\
   -x               run command once per task\n\
+  -X               implies -x and ignore calculator's stdout.\n\
   -y               magic line is used as an end-of-task marker\n\
+  -0               change paexec to expect NUL character as\n\
+                   a line separator instead of newline\n\
 \n\
   -r               include a node (or a number) to the output\n\
   -l               include 0-based task number to the output\n\
@@ -99,14 +104,18 @@ OPTIONS:\n\
                    of all tasks that depend on it,\n\
                    directly or indirectly\n\
 \n\
-  -m   s=<success>\n\
-       f=<failure>\n\
-       F=<fatal>\n\
-       t=<eot>\n\
-       d=<delimiter>\n\
-                   set an alternative for 'success',\n\
-                   'failure', 'fatal', '' (end-of-task marker) and\n\
-                   ' ' (task delimiter character).\n\
+  -m   s=<success> set an alternative for 'success' message\n\
+       f=<failure> set an alternative for 'failure' message\n\
+       F=<fatal>   set an alternative for 'fatal' message\n\
+       t=<EOT>     set an alternative for EOT marker\n\
+       d=<delimiter>    set the delimiter for -g mode.\n\
+                        The default is space character.\n\
+       w=<weight>  set an alternative for 'weight:' marker\n\
+\n\
+  -J replstr\n\
+                   execute command for each task, replacing\n\
+                   one or more occurrences of replstr with the entire task.\n\
+\n\
 -n and -c are mandatory options\n\
 \n\
 ");
@@ -169,7 +178,8 @@ static int end_of_stdin = 0;
 static const char *msg_success = "success";
 static const char *msg_failure = "failure";
 static const char *msg_fatal   = "fatal";
-static const char *msg_eot = "";
+static const char *msg_weight  = "weight:";
+static const char *msg_eot = NULL;
 char msg_delim = ' '; /* also used in tasks.c */
 
 static int resistant = 0;
@@ -182,6 +192,10 @@ static int exec_mode = 0;
 
 static int use_weights = 0;
 
+char eol_char = '\n';
+
+static char replstr[2] = "";
+
 struct envvar_entry {
 	SLIST_ENTRY(envvar_entry) entries;      /* List. */
 	char* name;
@@ -251,7 +265,7 @@ static void init__read_graph_tasks (void)
 	}
 
 	/* reading all tasks with their dependancies */
-	while (len = xgetline (&buf, &buf_sz, stdin), len != -1){
+	while (len = xgetdelim(&buf, &buf_sz, eol_char, stdin), len != -1){
 		if (len > 0 && buf [len-1] == '\n'){
 			buf [len-1] = 0;
 			--len;
@@ -291,7 +305,7 @@ static void init__read_graph_tasks (void)
 				break;
 		}
 
-		if (tok_cnt == 3 && !strcmp (tok1, "weight:")){
+		if (tok_cnt == 3 && !strcmp (tok1, msg_weight)){
 			/* weight: <task> <weight> */
 			id2 = tasks__add_task (xstrdup (tok2), atoi (tok3));
 			continue;
@@ -324,6 +338,36 @@ static void init__read_graph_tasks (void)
 	}
 }
 
+static char const runner_function1[] = "run() { %s \"$1\"; }";
+static char const runner_function2[] = "run() { %s; }";
+
+static char* generate_run_command(void)
+{
+	static char run_command [4096];
+	char *repl_ptr = NULL;
+
+	if (replstr[0]){
+		repl_ptr = arg_cmd;
+		while (repl_ptr = strstr(repl_ptr, replstr), repl_ptr != NULL){
+			repl_ptr[0] = '$';
+			repl_ptr[1] = '1';
+			repl_ptr += 2;
+		}
+
+		snprintf(run_command, sizeof(run_command), runner_function2, arg_cmd);
+	}else{
+		snprintf(run_command, sizeof(run_command), runner_function1, arg_cmd);
+	}
+
+	if (strlen(run_command) + 1 == sizeof(run_command)){
+		err_fatal ("paexec: Internal error6! (buffer size)");
+	}
+
+//	fprintf(stderr, "run_command: %s\n", run_command);
+
+	return run_command;
+}
+
 static void init__postproc_arg_cmd (void)
 {
 	char shq_cmd [4096];
@@ -348,13 +392,17 @@ static void init__postproc_arg_cmd (void)
 		}
 
 		snprintf (cmd, sizeof (cmd),
-				  "while read f; do"
-				  "  res=`%s \"$f\"`;"
+				  "%s\n while read f; do"
+				  "  res=`run \"$f\"`;"
 				  "  ex=$?;"
 				  "  %s" /* printing result */
 				  "  %s" /* condition. success/failure */
 				  "  echo '%s';" /* EOT */
-				  "done", arg_cmd, tmp, cond_cmd, magic_eot);
+				  "done", generate_run_command(), tmp, cond_cmd, magic_eot);
+
+		if (strlen(cmd) + 1 == sizeof(cmd)){
+			err_fatal ("paexec: Internal error7! (buffer size)");
+		}
 
 		xfree (arg_cmd);
 		arg_cmd = xstrdup (cmd);
@@ -967,20 +1015,30 @@ static void check_msg (const char *msg)
 static char *gen_cmd (int *argc, char ***argv)
 {
 	char cmd [4096];
+	char *curr_token;
 	size_t len;
 	size_t curr_len;
 	int i;
 
 	len = 0;
 	for (i=0; i < *argc; ++i){
-		curr_len = shquote ((*argv) [i], cmd+len, sizeof (cmd)-len-1);
+		curr_token = (*argv) [i];
+		if (replstr != NULL && !strcmp(curr_token, replstr)){
+			cmd[len+0] = '"';
+			cmd[len+1] = '$';
+			cmd[len+2] = '1';
+			cmd[len+3] = '"';
+			curr_len = 4;
+		}else{
+			curr_len = shquote (curr_token, cmd+len, sizeof (cmd)-len-1);
+		}
 		if (curr_len == (size_t)-1){
 			err_fatal ("paexec: Internal error4! (buffer size)");
 		}
 		len += curr_len;
 		cmd [len++] = ' ';
 
-		if (len >= sizeof (cmd)-1){
+		if (len >= sizeof (cmd)-20){ /* 20 chars is enough for "$1" :-) */
 			err_fatal ("paexec: Internal error5! (buffer size)");
 		}
 	}
@@ -997,7 +1055,7 @@ static void process_args (int *argc, char ***argv)
 	int mode_C = 0;
 
 	/* leading + is for shitty GNU libc */
-	static const char optstring [] = "+c:CdeEghiIlm:n:prst:VwW:xXyzZ:";
+	static const char optstring [] = "+0c:CdeEghiIJ:lm:n:prst:VwW:xXyzZ:";
 
 	while (c = getopt (*argc, *argv, optstring), c != EOF){
 		switch (c) {
@@ -1016,6 +1074,7 @@ static void process_args (int *argc, char ***argv)
 				assign_str (&arg_nodes, optarg);
 				break;
 			case 'c':
+				mode_C = 0;
 				arg_cmd = xstrdup (optarg);
 				break;
 			case 'C':
@@ -1049,6 +1108,20 @@ static void process_args (int *argc, char ***argv)
 				print_i2o = 1;
 				flush_i2o = 1;
 				break;
+			case 'J':
+				if (strlen(optarg) != 2){
+					fprintf(stderr, "-Jreplstr argument must have 2-characters length");
+					exit(2);
+				}
+
+				replstr[0] = optarg[0];
+				replstr[1] = optarg[1];
+
+				if (!msg_eot)
+					msg_eot = magic_eot;
+				if (!exec_mode)
+					exec_mode = 'x';
+				break;
 			case 's':
 			case 'g':
 				graph_mode = 1;
@@ -1076,10 +1149,12 @@ static void process_args (int *argc, char ***argv)
 				}else if (optarg [0] == 't' && optarg [1] == '='){
 					msg_eot = xstrdup (optarg+2);
 				}else if (optarg [0] == 'd' && optarg [1] == '='){
-					if (optarg [2] == 0 || optarg [3] != 0){
-						err_fatal ("paexec: bad argument for -md=. Exactly one character is allowed");
+					if (optarg [2] != 0 && optarg [3] != 0){
+						err_fatal ("paexec: bad argument for -md=. At most one character is allowed");
 					}
 					msg_delim = optarg [2];
+				}else if (optarg [0] == 'w' && optarg [1] == '='){
+					msg_weight = xstrdup (optarg+2);
 				}else{
 					err_fatal ("paexec: bad argument for -m");
 				}
@@ -1100,6 +1175,9 @@ static void process_args (int *argc, char ***argv)
 			case 'y':
 				msg_eot = magic_eot;
 				break;
+			case '0':
+				eol_char = '\0';
+				break;
 			default:
 				usage ();
 				exit (1);
@@ -1109,6 +1187,10 @@ static void process_args (int *argc, char ***argv)
 	*argv += optind;
 	*argc -= optind;
 
+	if (!msg_eot){
+		msg_eot = "";
+	}
+
 	if (mode_C){
 		if (!*argc){
 			err_fatal ("paexec: missing arguments. Run paexec -h for details");
